from typing import List

from pykube import Deployment, HTTPClient

from pytest_helm_charts.utils import wait_for_objects_condition


def _deployment_running(deploy: Deployment) -> bool:
    complete = (
        "status" in deploy.obj
        and "availableReplicas" in deploy.obj["status"]
        and "observedGeneration" in deploy.obj["status"]
        and "updatedReplicas" in deploy.obj["status"]
        and int(deploy.obj["status"]["observedGeneration"]) >= int(deploy.obj["metadata"]["generation"])
        and deploy.replicas == int(deploy.obj["status"]["updatedReplicas"])
        and deploy.replicas == int(deploy.obj["status"]["availableReplicas"])
        # Ensures the deployment has no unavailable replicas
        and ("unavailableReplicas" not in deploy.obj["status"] or 0 == int(deploy.obj["status"]["unavailableReplicas"]))
    )
    return complete


def wait_for_deployments_to_run(
    kube_client: HTTPClient,
    deployment_names: List[str],
    deployments_namespace: str,
    timeout_sec: int,
    missing_ok: bool = True,
) -> List[Deployment]:
    """
    Block until all the Deployments are running or timeout is reached.

    Args:
        kube_client: client to use to connect to the k8s cluster
        deployment_names: a list of Deployment names to check
        deployments_namespace: namespace where all the Deployments are created (single namespace for all resources)
        timeout_sec: timeout for the call
        missing_ok: when `True`, the function ignores that some of the objects listed in the `deployment_names`
            don't exist in k8s API and waits for them to show up; when `False`, a
            `pykube.exceptions.ObjectDoesNotExist` exception is raised.

    Returns:
        The list of Deployment resources with all the objects listed in `deployment_names` included.

    Raises:
        TimeoutError: when timeout is reached.
        pykube.exceptions.ObjectDoesNotExist: when `missing_ok == False` and one of the objects
            listed in `deployment_names` can't be found in k8s API

    """
    result = wait_for_objects_condition(
        kube_client,
        Deployment,
        deployment_names,
        deployments_namespace,
        _deployment_running,
        timeout_sec,
        missing_ok,
    )
    return result



"""This modules contains apps useful for testing HTTP applications."""

import math
from typing import Dict, Iterable, List, Optional, Protocol, Tuple

import pytest
from pykube import ConfigMap
from pytest_helm_charts.clusters import Cluster
from pytest_helm_charts.giantswarm_app_platform.app import AppFactoryFunc, ConfiguredApp
from pytest_helm_charts.utils import YamlDict, delete_and_wait_for_objects


class StormforgerLoadAppFactoryFunc(Protocol):
    def __call__(
        self,
        replicas: int,
        host_url: str,
        node_affinity_selector: Optional[Dict[str, str]] = None,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
    ) -> ConfiguredApp: ...


@pytest.fixture(scope="module")
def stormforger_load_app_factory(app_factory: AppFactoryFunc) -> StormforgerLoadAppFactoryFunc:
    """A factory fixture to return a function that can produce Stromforger Load App instances.
    Fixture's scope is 'module'..

    Args:
        app_factory: auto-injected [app_factory](pytest_helm_charts.giantswarm_app_platform.app.app_factory) fixture.

    Returns:
        A function you can use to create stormforger instances. The function has the following args.

    Examples:
        Create and run using 8 replicas erving the 'loadtest.local' URL. Use affinity selector to run
        on the 'localhost' Kubernetes Node.

        >>> stormforger_load_app_factory(8, "loadtest.local", {"kubernetes.io/hostname": "localhost"})
    """

    def _stormforger_load_app_factory(
        replicas: int,
        host_url: str,
        node_affinity_selector: Optional[Dict[str, str]] = None,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
    ) -> ConfiguredApp:
        """Creates and deploys stormforger load app by creating the relevant App CR in the API.

        Args:
            replicas: number of replicas of Pods the app should run.
            host_url: the URL under which the app will serve requests.
            node_affinity_selector: option node affinity Kubernetes selector. Default={default}.
             extra_metadata: optional dict that will be merged with the 'metadata:' section of the object
             extra_spec: optional dict that will be merged with the 'spec:' section of the object

        Returns:
            [AppCR](AppCR) describing the CR created in API.

        """
        config_values: YamlDict = {
            "replicaCount": replicas,
            "ingress": {
                "enabled": "true",
                "annotations": {"kubernetes.io/ingress.class": "nginx"},
                "paths": ["/"],
                "hosts": [host_url],
            },
            "autoscaling": {"enabled": "false"},
        }

        if node_affinity_selector is not None:
            config_values["nodeAffinity"] = {"enabled": "true", "selector": node_affinity_selector}
        stormforger_app = app_factory(
            "loadtest-app",
            "0.2.0",
            "default",
            "default",
            "https://giantswarm.github.io/default-catalog/",
            config_values=config_values,
            extra_metadata=extra_metadata,
            extra_spec=extra_spec,
        )
        return stormforger_app

    return _stormforger_load_app_factory


class GatlingAppFactoryFunc(Protocol):
    def __call__(
        self,
        simulation_file: str,
        node_affinity_selector: Optional[Dict[str, str]] = None,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
    ) -> ConfiguredApp: ...


@pytest.fixture(scope="module")
def gatling_app_factory(kube_cluster: Cluster, app_factory: AppFactoryFunc) -> Iterable[GatlingAppFactoryFunc]:
    """A factory fixture to return a function that can produce Gatling instances. Gatling is a HTTP
    performance testing tool. Fixture's scope is 'module'..

    Args:
        app_factory: auto-injected [app_factory](pytest_helm_charts.giantswarm_app_platform.app.app_factory) fixture.
        kube_cluster: auto-injected [kube_cluster](pytest_helm_charts.fixtures.kube_cluster) fixture.

    Returns:
        A function you can use to create Gatling instances.
    """

    created_configmaps: List[ConfigMap] = []

    def _gatling_app_factory(
        simulation_file: str,
        node_affinity_selector: Optional[Dict[str, str]] = None,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
    ) -> ConfiguredApp:
        namespace = "default"
        with open(simulation_file) as f:
            simulation_code = f.read()
        simulation_cm: YamlDict = {
            "apiVersion": "v1",
            "kind": "ConfigMap",
            "metadata": {"name": "gatling-simulation", "namespace": namespace, "labels": {"app": "gatling"}},
            "data": {"NginxSimulation.scala": simulation_code},
        }

        config_values: YamlDict = {
            "simulation": {"configMap": {"name": simulation_cm["metadata"]["name"]}, "name": "nginx.NginxSimulation"}
        }

        if node_affinity_selector is not None:
            config_values["nodeAffinity"] = {"enabled": "true", "selector": node_affinity_selector}

        simulation_cm_obj = ConfigMap(kube_cluster.kube_client, simulation_cm)
        simulation_cm_obj.create()
        created_configmaps.append(simulation_cm_obj)
        gatling_app = app_factory(
            "gatling-app",
            "1.0.2",
            "giantswarm-playground",
            "default",
            "https://giantswarm.github.io/giantswarm-playground-catalog/",
            namespace=namespace,
            config_values=config_values,
            extra_metadata=extra_metadata,
            extra_spec=extra_spec,
        )
        return gatling_app

    yield _gatling_app_factory

    delete_and_wait_for_objects(kube_cluster.kube_client, ConfigMap, created_configmaps)


class GatlingParser:
    def __init__(self, lines: str) -> None:
        split_lines = lines.splitlines()
        start_line_query = [i for i in range(len(split_lines)) if split_lines[i] == "Generating reports..."]
        assert len(start_line_query) == 1
        start_line = start_line_query[0]
        report_lines = split_lines[start_line + 2 : start_line + 20]
        assert report_lines[0] == "================================================================================"
        assert report_lines[1] == "---- Global Information --------------------------------------------------------"

        total, ok, bad = self.__parse_result_line(report_lines[2], "request count")
        self.request_count_total, self.request_count_ok, self.request_count_bad = int(total), int(ok), int(bad)
        self.request_success_ratio = float(self.request_count_ok) / float(self.request_count_total)
        self.request_failure = float(self.request_count_bad) / float(self.request_count_total)

        min_time, _, _ = self.__parse_result_line(report_lines[3], "min response time")
        self.min_response_time = int(min_time)

        max_time, _, _ = self.__parse_result_line(report_lines[4], "max response time")
        self.max_response_time = int(max_time)

        mean_time, _, _ = self.__parse_result_line(report_lines[5], "mean response time")
        self.mean_response_time = int(mean_time)

        std_deviation, _, _ = self.__parse_result_line(report_lines[6], "std deviation")
        self.std_deviation = int(std_deviation)

        response_time_50th_percentile, _, _ = self.__parse_result_line(report_lines[7], "response time 50th percentile")
        self.response_time_50th_percentile = int(response_time_50th_percentile)

        response_time_75th_percentile, _, _ = self.__parse_result_line(report_lines[8], "response time 75th percentile")
        self.response_time_75th_percentile = int(response_time_75th_percentile)

        response_time_95th_percentile, _, _ = self.__parse_result_line(report_lines[9], "response time 95th percentile")
        self.response_time_95th_percentile = int(response_time_95th_percentile)

        response_time_99th_percentile, _, _ = self.__parse_result_line(
            report_lines[10], "response time 99th percentile"
        )
        self.response_time_99th_percentile = int(response_time_99th_percentile)

        mean_rps, _, _ = self.__parse_result_line(report_lines[11], "mean requests/sec")
        self.mean_rps = float(mean_rps)

        assert report_lines[12] == "---- Response Time Distribution ------------------------------------------------"
        self.response_time_distribution: Dict[Tuple[float, float], int] = {}

        line = report_lines[13][2:]
        fields = list(filter(None, line.split(" ")))
        assert fields[0] == "t"
        upper = float(fields[2])
        self.response_time_distribution[(0, upper)] = int(fields[4])

        line = report_lines[14][2:]
        fields = list(filter(None, line.split(" ")))
        assert fields[3] == "t"
        lower = float(fields[0])
        upper = float(fields[5])
        self.response_time_distribution[(lower, upper)] = int(fields[7])

        line = report_lines[15][2:]
        fields = list(filter(None, line.split(" ")))
        assert fields[0] == "t"
        lower = float(fields[2])
        self.response_time_distribution[(lower, math.inf)] = int(fields[4])

        line = report_lines[16][2:]
        fields = list(filter(None, line.split(" ")))
        assert fields[0] == "failed"

    @staticmethod
    def __parse_result_line(line: str, header: str) -> Tuple[str, str, str]:
        assert line[0:2] == "> "
        line = line[2:]
        assert line.find(header) >= 0
        fields = line[len(header) :].split(" ")
        fields = list(filter(None, fields))
        total = fields[0]
        ok = fields[1].split("=")[1]
        bad = fields[2].split("=")[1].rstrip(")")
        return total, ok, bad



"""This module introduces classes for handling different clusters."""

import json
import logging
import shutil
import subprocess  # nosec
from abc import ABC, abstractmethod
from typing import Optional, Any

from pykube import HTTPClient, KubeConfig

logger = logging.getLogger(__name__)


class Cluster(ABC):
    """Represents an abstract cluster."""

    _kube_client: Optional[HTTPClient]
    kube_config_path: Optional[str]

    def __init__(self, kube_config_path: Optional[str] = None):
        super().__init__()
        self._kube_client = None
        self.kube_config_path = kube_config_path

    @abstractmethod
    def create(self) -> HTTPClient:
        """Creates an instance of a cluster and returns HTTPClient to connect to it."""
        raise NotImplementedError

    @abstractmethod
    def destroy(self) -> None:
        """Destroys the cluster created earlier with a call to [create](Cluster.create)."""
        raise NotImplementedError

    @property
    def kube_client(self) -> Optional[HTTPClient]:
        """Returns the HTTP client you can use to access Kubernetes API of the cluster under test.

        Please refer to [pykube](https://pykube.readthedocs.io/en/latest/api/pykube.html) to get docs
        for [HTTPClient](https://pykube.readthedocs.io/en/latest/api/pykube.html#pykube.http.HTTPClient).
        """
        return self._kube_client

    def kubectl(  # noqa: C901
        self,
        subcmd_string: str,
        std_input: str = "",
        output_format: str = "json",
        use_shell: bool = False,
        **kwargs: str,
    ) -> Any:
        """Execute command by running 'kubectl' binary.

        If your cluster delivers the kube.config file, run a kubectl command against the cluster
        and return the output. Otherwise, exception is raised.

        Args:
            subcmd_string (str): Command to run, like "delete pod abc"
            std_input (str): Use this to pass a manifest file directly as a string (results in 'kubectl [cmd] -f -')
            output_format (str): Option "--output" as passed to 'kubectl'. Default is 'json', make sure to change to ""
                for commands that don't return JSON.
            use_shell: Whether the 'kubectl' command should be invoked directly (when 'False') or wrapped in system
                shell ('True'). 'False' by default.
            kwargs: arbitrary dictionary of options and values that will be passed directly to 'kubectl'

        Returns:
            str: The output printed by 'kubectl', if the command succeeded (exit code was 0)

        Raises:
            subprocess.CalledProcessError: If the command exited with non-zero exit code
        """
        if not self.kube_config_path:
            raise ValueError("'kube_config_path' can't be empty to use 'kubectl'")
        bin_name = "kubectl"
        if shutil.which(bin_name) is None:
            raise ValueError(f"Can't find {bin_name} executable. Please make sure it's available in $PATH.")

        subcmds = subcmd_string.split(" ")
        if len(subcmds) == 0:
            raise ValueError("You need to give at least one kubectl subcommand in the subcmd_string argument.")

        kwargs["kubeconfig"] = self.kube_config_path
        if std_input:
            kwargs["filename"] = "-"
        if output_format:
            kwargs["output"] = output_format
        if subcmds[0].lower() in ["delete", "patch", "label", "annotate"] and "output" in kwargs:
            del kwargs["output"]

        options = {f"--{option}={value}" for option, value in sorted(kwargs.items())}

        try:
            cmd = (
                bin_name + " " + subcmd_string + " " + " ".join(options)
                if use_shell
                else [bin_name, *subcmds, *options]
            )
            result = subprocess.check_output(
                cmd,
                stderr=subprocess.PIPE,
                encoding="utf-8",
                input=std_input,
                shell=use_shell,  # nosec
            )
        except subprocess.CalledProcessError as e:
            logger.error(
                f"'kubectl' call returned an error. Exit code: '{e.returncode}', stdout: '{e.stdout}',"
                f"stderr: '{e.stderr}'"
            )
            raise

        # return result from kubectl command:
        # - as parsed json by default, extracting objects list to list
        # - as plain text otherwise

        if "output" in kwargs and kwargs["output"] == "json":
            output_json = json.loads(result)
            if "items" in output_json:
                return output_json["items"]
            return output_json
        return result


class ExistingCluster(Cluster):
    """Implementation of [Cluster](Cluster) that uses kube.config file to connect to external
    existing cluster.
    """

    def __init__(self, kube_config_path: str) -> None:
        super().__init__(kube_config_path)

    def create(self) -> HTTPClient:
        kube_config = KubeConfig.from_file(self.kube_config_path)
        self._kube_client = HTTPClient(kube_config)
        return self._kube_client

    def destroy(self) -> None:
        if self._kube_client is None:
            return
        self._kube_client.session.close()
        self._kube_client = None



import logging
from dataclasses import dataclass, field, asdict
from typing import Protocol, Optional, List, Any, Dict

from pykube import HTTPClient

from pytest_helm_charts.k8s.fixtures import NamespaceFactoryFunc
from pytest_helm_charts.flux.utils import NamespacedFluxCR, flux_cr_ready
from pytest_helm_charts.utils import wait_for_objects_condition, inject_extra


logger = logging.getLogger(__name__)


class HelmReleaseCR(NamespacedFluxCR):
    version = "helm.toolkit.fluxcd.io/v2beta1"
    endpoint = "helmreleases"
    kind = "HelmRelease"


@dataclass
class CrossNamespaceObjectReference:
    kind: str
    name: str
    apiVersion: Optional[str] = None
    namespace: Optional[str] = None


@dataclass
class ChartTemplate:
    chart: str
    sourceRef: CrossNamespaceObjectReference
    version: Optional[str] = None
    valuesFiles: List[str] = field(default_factory=list)


@dataclass
class ValuesReference:
    kind: str
    name: str
    valuesKey: Optional[str] = None
    targetPath: Optional[str] = None
    optional: Optional[bool] = None


class HelmReleaseFactoryFunc(Protocol):
    def __call__(
        self,
        name: str,
        namespace: str,
        chart: ChartTemplate,
        interval: str,
        suspend: bool = False,
        release_name: Optional[str] = None,
        target_namespace: Optional[str] = None,
        depends_on: Optional[List[CrossNamespaceObjectReference]] = None,
        timeout: Optional[str] = None,
        values_from: Optional[List[ValuesReference]] = None,
        values: Optional[dict] = None,
        service_account_name: Optional[str] = None,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
        wait_timeout_sec: int = 30,
    ) -> HelmReleaseCR: ...


def helm_release_factory_func(
    kube_client: HTTPClient,
    namespace_factory: NamespaceFactoryFunc,
    created_helm_releases: List[HelmReleaseCR],
) -> HelmReleaseFactoryFunc:
    """Return a factory object, that can be used to create a new HelmRelease CRs"""

    def _helm_release_factory(
        name: str,
        namespace: str,
        chart: ChartTemplate,
        interval: str,
        suspend: bool = False,
        release_name: Optional[str] = None,
        target_namespace: Optional[str] = None,
        depends_on: Optional[List[CrossNamespaceObjectReference]] = None,
        timeout: Optional[str] = None,
        values_from: Optional[List[ValuesReference]] = None,
        values: Optional[dict] = None,
        service_account_name: Optional[str] = None,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
        wait_timeout_sec: int = 30,
    ) -> HelmReleaseCR:
        """A factory function used to create Flux HelmRepository.
        Args:
            name: name of the created Kustomization CR.
            namespace: namespace to create the Kustomization CR in.
            chart: Chart defines the template that should be created for this HelmRelease.
            interval: the interval at which to check for repository updates.
            suspend: This flag tells the controller to suspend the reconciliation of this source.
            release_name: Name used for the Helm release.
            target_namespace: Namespace to target when performing operations for the HelmRelease.
            depends_on: a list of CrossNamespaceObjectReference with
                references to HelmRelease resources that must be ready before this HelmRelease.
            timeout: The timeout of index downloading, defaults to 60s.
            wait_timeout_sec: How long to wait for the HelmRelease to be ready.
            values_from: References to resources containing Helm values for this HelmRelease,
                and information about how they should be merged.
            values: Holds the values for this Helm release.
            service_account_name: The name of the Kubernetes service account to impersonate
                when reconciling this HelmRelease.
            extra_metadata: a dictionary of any additional attributes to put directly into "metadata"
                part of the object
            extra_spec: a dictionary of any additional attributes to put directly into "spec"
                part of the object
            wait_timeout_sec: How long to wait for the HelmRelease to be ready.
        Returns:
            HelmRelease created or found in the k8s API.
        Raises:
            ValueError: if object with the same name already exists.
        """
        for hr in created_helm_releases:
            if hr.metadata["name"] == name and hr.metadata["namespace"] == namespace:
                return hr

        namespace_factory(namespace)
        if target_namespace:
            namespace_factory(target_namespace)
        helm_release = make_helm_release_obj(
            kube_client,
            name,
            namespace,
            chart,
            interval,
            suspend,
            release_name,
            target_namespace,
            depends_on,
            timeout,
            values_from,
            values,
            service_account_name,
            extra_metadata=extra_metadata,
            extra_spec=extra_spec,
        )
        created_helm_releases.append(helm_release)
        helm_release.create()
        logger.debug(f"Created Flux HelmRelease '{helm_release.namespace}/{helm_release.name}'.")
        wait_for_helm_releases_to_be_ready(kube_client, [name], namespace, wait_timeout_sec, missing_ok=True)
        return helm_release

    return _helm_release_factory


def make_helm_release_obj(
    kube_client: HTTPClient,
    name: str,
    namespace: str,
    chart: ChartTemplate,
    interval: str,
    suspend: bool = False,
    release_name: Optional[str] = None,
    target_namespace: Optional[str] = None,
    depends_on: Optional[List[CrossNamespaceObjectReference]] = None,
    timeout: Optional[str] = None,
    values_from: Optional[List[ValuesReference]] = None,
    values: Optional[dict] = None,
    service_account_name: Optional[str] = None,
    extra_metadata: Optional[dict] = None,
    extra_spec: Optional[dict] = None,
) -> HelmReleaseCR:
    cr: Dict[str, Any] = {
        "apiVersion": HelmReleaseCR.version,
        "kind": HelmReleaseCR.kind,
        "metadata": {
            "name": name,
            "namespace": namespace,
        },
        "spec": {
            "chart": {"spec": asdict(chart)},
            "interval": interval,
            "suspend": suspend,
        },
    }
    if release_name:
        cr["spec"]["releaseName"] = release_name
    if target_namespace:
        cr["spec"]["targetNamespace"] = target_namespace
    if depends_on:
        cr["spec"]["dependsOn"] = [asdict(d) for d in depends_on]
    if timeout:
        cr["spec"]["timeout"] = timeout
    if values_from:
        cr["spec"]["valuesFrom"] = [asdict(v) for v in values_from]
    if values:
        cr["spec"]["values"] = values
    if service_account_name:
        cr["spec"]["serviceAccountName"] = service_account_name

    cr = inject_extra(cr, extra_metadata, extra_spec)
    return HelmReleaseCR(kube_client, cr)


def wait_for_helm_releases_to_be_ready(
    kube_client: HTTPClient,
    helm_release_names: List[str],
    helm_release_namespace: str,
    timeout_sec: int,
    missing_ok: bool = False,
) -> List[HelmReleaseCR]:
    """Block until all Helm Release objects in `helm_release_names` have status 'Ready'."""
    objects = wait_for_objects_condition(
        kube_client,
        HelmReleaseCR,
        helm_release_names,
        helm_release_namespace,
        flux_cr_ready,
        timeout_sec,
        missing_ok,
    )
    return objects



from typing import Iterable, List

import pykube
import pytest

from pytest_helm_charts.k8s.deployment import wait_for_deployments_to_run
from pytest_helm_charts.k8s.fixtures import NamespaceFactoryFunc
from pytest_helm_charts.clusters import Cluster
from pytest_helm_charts.flux.git_repository import (
    GitRepositoryCR,
    GitRepositoryFactoryFunc,
    git_repository_factory_func,
)
from pytest_helm_charts.flux.helm_release import HelmReleaseCR, HelmReleaseFactoryFunc, helm_release_factory_func
from pytest_helm_charts.flux.helm_repository import (
    HelmRepositoryCR,
    HelmRepositoryFactoryFunc,
    helm_repository_factory_func,
)
from pytest_helm_charts.flux.kustomization import KustomizationCR, KustomizationFactoryFunc, kustomization_factory_func
from pytest_helm_charts.utils import delete_and_wait_for_objects

FLUX_NAMESPACE_NAME = "default"
FLUX_DEPLOYMENTS_READY_TIMEOUT: int = 180


@pytest.fixture(scope="module")
def flux_deployments(kube_cluster: Cluster) -> List[pykube.Deployment]:
    """
    Waits (blocks) until all the Deployments used to deploy Flux itself run
    (a.k.a. "wait for Flux to be ready for use").

    Args:
        kube_cluster: Cluster to use to connect to the k8s cluster

    Returns: A List of `pykube.Deployment` objects used by Flux itself.

    """
    deployments = wait_for_deployments_to_run(
        kube_cluster.kube_client,
        [
            "helm-controller",
            "image-automation-controller",
            "image-reflector-controller",
            "kustomize-controller",
            "notification-controller",
            "source-controller",
        ],
        FLUX_NAMESPACE_NAME,
        FLUX_DEPLOYMENTS_READY_TIMEOUT,
    )
    return deployments


@pytest.fixture(scope="function")
def kustomization_factory_function_scope(
    kube_cluster: Cluster, namespace_factory: NamespaceFactoryFunc
) -> Iterable[KustomizationFactoryFunc]:
    """Returns function-scoped [Kustomization](https://fluxcd.io/docs/components/kustomize/kustomization/)
    factory."""
    yield from _kustomization_factory_impl(kube_cluster, namespace_factory)


@pytest.fixture(scope="module")
def kustomization_factory(
    kube_cluster: Cluster, namespace_factory: NamespaceFactoryFunc
) -> Iterable[KustomizationFactoryFunc]:
    """Returns module-scoped [Kustomization](https://fluxcd.io/docs/components/kustomize/kustomization/)
    factory."""
    yield from _kustomization_factory_impl(kube_cluster, namespace_factory)


def _kustomization_factory_impl(
    kube_cluster: Cluster, namespace_factory: NamespaceFactoryFunc
) -> Iterable[KustomizationFactoryFunc]:
    created_objects: List[KustomizationCR] = []

    yield kustomization_factory_func(kube_cluster.kube_client, namespace_factory, created_objects)

    delete_and_wait_for_objects(kube_cluster.kube_client, KustomizationCR, created_objects)


@pytest.fixture(scope="function")
def git_repository_factory_function_scope(
    kube_cluster: Cluster, namespace_factory: NamespaceFactoryFunc
) -> Iterable[GitRepositoryFactoryFunc]:
    """Returns function-scoped [Git Repository](https://fluxcd.io/docs/components/source/gitrepositories/) factory."""
    yield from _git_repository_factory_impl(kube_cluster, namespace_factory)


@pytest.fixture(scope="module")
def git_repository_factory(
    kube_cluster: Cluster, namespace_factory: NamespaceFactoryFunc
) -> Iterable[GitRepositoryFactoryFunc]:
    """Returns module-scoped [Git Repository](https://fluxcd.io/docs/components/source/gitrepositories/) factory."""
    yield from _git_repository_factory_impl(kube_cluster, namespace_factory)


def _git_repository_factory_impl(
    kube_cluster: Cluster, namespace_factory: NamespaceFactoryFunc
) -> Iterable[GitRepositoryFactoryFunc]:
    created_objects: List[GitRepositoryCR] = []

    yield git_repository_factory_func(kube_cluster.kube_client, namespace_factory, created_objects)

    delete_and_wait_for_objects(kube_cluster.kube_client, GitRepositoryCR, created_objects)


@pytest.fixture(scope="function")
def helm_repository_factory_function_scope(
    kube_cluster: Cluster, namespace_factory: NamespaceFactoryFunc
) -> Iterable[HelmRepositoryFactoryFunc]:
    """Returns function-scoped [Helm Repository](https://fluxcd.io/docs/components/source/helmrepositories/) factory."""
    yield from _helm_repository_factory_impl(kube_cluster, namespace_factory)


@pytest.fixture(scope="module")
def helm_repository_factory(
    kube_cluster: Cluster, namespace_factory: NamespaceFactoryFunc
) -> Iterable[HelmRepositoryFactoryFunc]:
    """Returns module-scoped [Helm Repository](https://fluxcd.io/docs/components/source/helmrepositories/) factory."""
    yield from _helm_repository_factory_impl(kube_cluster, namespace_factory)


def _helm_repository_factory_impl(
    kube_cluster: Cluster, namespace_factory: NamespaceFactoryFunc
) -> Iterable[HelmRepositoryFactoryFunc]:
    created_objects: List[HelmRepositoryCR] = []

    yield helm_repository_factory_func(kube_cluster.kube_client, namespace_factory, created_objects)

    delete_and_wait_for_objects(kube_cluster.kube_client, HelmRepositoryCR, created_objects)


@pytest.fixture(scope="function")
def helm_release_factory_function_scope(
    kube_cluster: Cluster, namespace_factory: NamespaceFactoryFunc
) -> Iterable[HelmReleaseFactoryFunc]:
    """Returns function-scoped [Helm Release](https://fluxcd.io/docs/components/helm/helmreleases/) factory."""
    yield from _helm_release_factory_impl(kube_cluster, namespace_factory)


@pytest.fixture(scope="module")
def helm_release_factory(
    kube_cluster: Cluster, namespace_factory: NamespaceFactoryFunc
) -> Iterable[HelmReleaseFactoryFunc]:
    """Returns module-scoped [Helm Release](https://fluxcd.io/docs/components/helm/helmreleases/) factory."""
    yield from _helm_release_factory_impl(kube_cluster, namespace_factory)


def _helm_release_factory_impl(
    kube_cluster: Cluster, namespace_factory: NamespaceFactoryFunc
) -> Iterable[HelmReleaseFactoryFunc]:
    created_objects: List[HelmReleaseCR] = []

    yield helm_release_factory_func(kube_cluster.kube_client, namespace_factory, created_objects)

    delete_and_wait_for_objects(kube_cluster.kube_client, HelmReleaseCR, created_objects)



import logging
from copy import deepcopy
from typing import List, Protocol, Optional, NamedTuple

import pykube
import yaml
from pykube import HTTPClient, ConfigMap
from pykube.objects import NamespacedAPIObject

from pytest_helm_charts.k8s.fixtures import NamespaceFactoryFunc
from pytest_helm_charts.giantswarm_app_platform.catalog import CatalogFactoryFunc
from pytest_helm_charts.utils import YamlDict, wait_for_objects_condition, inject_extra


logger = logging.getLogger(__name__)


class AppCR(NamespacedAPIObject):
    version = "application.giantswarm.io/v1alpha1"
    endpoint = "apps"
    kind = "App"


class ConfiguredApp(NamedTuple):
    """Class that represents application deployed by App CR and its optional configuration in ConfigMap."""

    app: AppCR
    app_cm: Optional[ConfigMap]


class AppFactoryFunc(Protocol):
    def __call__(
        self,
        app_name: str,
        app_version: str,
        catalog_name: str,
        catalog_namespace: str,
        catalog_url: str,
        namespace: str = "default",
        deployment_namespace: str = "default",
        config_values: Optional[YamlDict] = None,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
        timeout_sec: int = 60,
    ) -> ConfiguredApp: ...


def app_factory_func(
    kube_client: HTTPClient,
    catalog_factory: CatalogFactoryFunc,
    namespace_factory: NamespaceFactoryFunc,
    created_apps: List[ConfiguredApp],
) -> AppFactoryFunc:
    def _app_factory(
        app_name: str,
        app_version: str,
        catalog_name: str,
        catalog_namespace: str,
        catalog_url: str,
        namespace: str = "default",
        deployment_namespace: str = "default",
        config_values: Optional[YamlDict] = None,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
        timeout_sec: int = 60,
    ) -> ConfiguredApp:
        """Factory function used to create and deploy new apps using App CR. Calls are blocking.

        Args:
             app_name: name of the app in the app catalog
             app_version: version of the app to use from the app catalog
             catalog_name: a name of the catalog used for the
                [CatalogCR](pytest_helm_charts.giantswarm_app_platform.catalog.CatalogCR);
                new catalog is created only when one with the same name doesn't already exist
             catalog_namespace: namespace where the catalog catalog_name is defined
             catalog_url: URL of the catalog to install the application from; this is used only if a catalog
                with the same name doesn't already exist (then a new catalog with the given name and URL is created
                in the k8s API)
             namespace: namespace where the App CR will be created
             deployment_namespace: namespace where the app will be deployed (can be different from `namespace`)
             config_values: any values that should be used to configure the app (same as `values.yaml` used for
                a Helm Chart directly).
             extra_metadata: optional dict that will be merged with the 'metadata:' section of the object
             extra_spec: optional dict that will be merged with the 'spec:' section of the object
             timeout_sec: timeout in seconds for the create operation

        Returns:
            The [ConfiguredApp](ConfiguredApp) object that includes both AppCR and ConfigMap created to
            deploy the app.

        Raises:
            pykube.exceptions.ObjectDoesNotExist: if for any reason the created App CR object doesn't exist after
                creation and it's impossible to check its readiness.
                TimeoutError: when the timeout has been reached.
        """
        assert catalog_url != ""
        catalog_factory(catalog_name, catalog_namespace, catalog_url)
        namespace_factory(namespace)
        configured_app = create_app(
            kube_client,
            app_name,
            app_version,
            catalog_name,
            catalog_namespace,
            namespace,
            deployment_namespace,
            config_values,
            extra_metadata,
            extra_spec,
        )
        created_apps.append(configured_app)
        logger.debug(f"Created App '{configured_app.app.namespace}/{configured_app.app.name}'.")
        if timeout_sec > 0:
            wait_for_apps_to_run(kube_client, [app_name], namespace, timeout_sec)

        # we return a new object here, so that user doesn't alter the one added to created_apps
        return deepcopy(configured_app)

    return _app_factory


def _app_has_status(app: AppCR, status: str) -> bool:
    complete = (
        "status" in app.obj
        and "release" in app.obj["status"]
        and "appVersion" in app.obj["status"]
        and "status" in app.obj["status"]["release"]
        and app.obj["status"]["release"]["status"].lower() == status
    )
    return complete


def _app_failed(app: AppCR) -> bool:
    return _app_has_status(app, "failed")


def _app_deployed(app: AppCR) -> bool:
    return _app_has_status(app, "deployed")


def _app_deleted(app: AppCR) -> bool:
    return _app_has_status(app, "deleted")


def wait_for_apps_to_run(
    kube_client: HTTPClient,
    app_names: List[str],
    app_namespace: str,
    timeout_sec: int,
    missing_ok: bool = False,
    fail_fast: bool = False,
) -> List[AppCR]:
    """
    Block until all the apps are running or timeout is reached.

    Args:
        kube_client: client to use to connect to the k8s cluster
        app_names: a list of application names to check; all of the applications must be running for this
            function to succeed
        app_namespace: namespace where the App CRs of all the apps are stored
        timeout_sec: timeout for the call
        missing_ok: when `True`, the function ignores that some of the apps listed in the `app_names`
            don't exist in k8s API and waits for them to show up; when `False`, an
            [ObjectNotFound](pykube.exceptions.ObjectDoesNotExist) exception is raised.
        fail_fast: if set to True, the function fails as soon as the App reaches 'status=failed`, without
            waiting for any subsequent status changes.

    Returns:
        The list of App CRs with all the apps listed in `app_names` included.

    Raises:
        TimeoutError: when timeout is reached.
        pykube.exceptions.ObjectDoesNotExist: when `missing_ok == False` and one of the apps
            listed in `app_names` can't be found in k8s API
        ObjectStatusError: when App object has `Status: failed` status.

    """
    apps = wait_for_objects_condition(
        kube_client,
        AppCR,
        app_names,
        app_namespace,
        _app_deployed,
        timeout_sec,
        missing_ok,
        _app_failed if fail_fast else None,
    )
    return apps


def wait_for_app_to_be_deleted(
    kube_client: HTTPClient,
    app_name: str,
    app_namespace: str,
    timeout_sec: int,
) -> bool:
    """
    Block until an App CR has status `deleted` or doesn't exist in k8s API.

    Args:
        kube_client: client to use to connect to the k8s cluster
        app_name: an application name to check
        app_namespace: namespace where all the App CRs are stored
        timeout_sec: timeout for the call

    Returns:
        `True` when the App CR was found with status `deleted` or was not found at all. `False` otherwise.

    Raises:
        TimeoutError: when timeout is reached.

    """
    try:
        apps = wait_for_objects_condition(
            kube_client, AppCR, [app_name], app_namespace, _app_deleted, timeout_sec, missing_ok=False
        )
    except pykube.exceptions.ObjectDoesNotExist:
        return True
    return len(apps) == 1


def delete_app(configured_app: ConfiguredApp) -> None:
    """
    Deletes the app created by [create_app](create_app).
    Args:
        configured_app: ConfiguredApp (with optional ConfigMap configuration) to be deleted.

    Returns:
        None
    """
    configured_app.app.delete()
    if configured_app.app_cm:
        configured_app.app_cm.delete()


def make_app_object(
    kube_client: HTTPClient,
    app_name: str,
    app_version: str,
    catalog_name: str,
    catalog_namespace: str,
    namespace: str,
    deployment_namespace: str,
    config_values: Optional[YamlDict] = None,
    extra_metadata: Optional[dict] = None,
    extra_spec: Optional[dict] = None,
) -> ConfiguredApp:
    """Creates a new App object. Optionally creates a values ConfigMap. Objects are not sent to API server.

    Args:
        kube_client: client to use to connect to the k8s cluster
        app_name: name of the app in the app catalog
        app_version: version of the app to use from the app catalog
        catalog_name: a name of the catalog used for the
            [CatalogCR](pytest_helm_charts.giantswarm_app_platform.catalog.CatalogCR);
            must already exist
        catalog_namespace: a namespace of the
            [CatalogCR](pytest_helm_charts.giantswarm_app_platform.catalog.CatalogCR)
        namespace: namespace where the App CR will be created
        deployment_namespace: namespace where the app will be deployed (can be different than `namespace`)
        config_values: any values that should be used to configure the app (same as `values.yaml` used for
            a Helm Chart directly).
        extra_metadata: optional dict that will be merged with the 'metadata:' section of the object
        extra_spec: optional dict that will be merged with the 'spec:' section of the object

    Returns:
        The [ConfiguredApp](ConfiguredApp) object that includes both AppCR and ConfigMap.
    """
    if config_values is None:
        config_values = {}

    # TODO: include proper regexp validation
    assert app_name != ""
    assert app_version != ""
    assert catalog_name != ""
    api_version = "application.giantswarm.io/v1alpha1"
    app_cm_name = "{}-testing-user-config".format(app_name)
    kind = "App"
    app: YamlDict = {
        "apiVersion": api_version,
        "kind": kind,
        "metadata": {
            "name": app_name,
            "namespace": namespace,
            "labels": {"app": app_name, "app-operator.giantswarm.io/version": "0.0.0"},
        },
        "spec": {
            "catalog": catalog_name,
            "catalogNamespace": catalog_namespace,
            "version": app_version,
            "kubeConfig": {"inCluster": True},
            "name": app_name,
            "namespace": deployment_namespace,
        },
    }
    app_cm_obj: Optional[ConfigMap] = None
    if config_values:
        app["spec"]["config"] = {"configMap": {"name": app_cm_name, "namespace": namespace}}
        app_cm: YamlDict = {
            "apiVersion": "v1",
            "kind": "ConfigMap",
            "metadata": {"name": app_cm_name, "namespace": namespace},
            "data": {"values": yaml.dump(config_values)},
        }
        app_cm_obj = ConfigMap(kube_client, app_cm)
    app = inject_extra(app, extra_metadata, extra_spec)
    app_obj = AppCR(kube_client, app)
    return ConfiguredApp(app_obj, app_cm_obj)


def create_app(
    kube_client: HTTPClient,
    app_name: str,
    app_version: str,
    catalog_name: str,
    catalog_namespace: str,
    namespace: str,
    deployment_namespace: str,
    config_values: Optional[YamlDict] = None,
    extra_metadata: Optional[dict] = None,
    extra_spec: Optional[dict] = None,
) -> ConfiguredApp:
    configured_app = make_app_object(
        kube_client,
        app_name,
        app_version,
        catalog_name,
        catalog_namespace,
        namespace,
        deployment_namespace,
        config_values,
        extra_metadata,
        extra_spec,
    )
    if configured_app.app_cm:
        configured_app.app_cm.create()
    configured_app.app.create()
    return configured_app



import random
import string
from typing import Protocol, Optional, Iterable, List

import pykube
import pytest

from pytest_helm_charts.clusters import Cluster
from pytest_helm_charts.fixtures import logger
from pytest_helm_charts.k8s.namespace import ensure_namespace_exists
from pytest_helm_charts.utils import delete_and_wait_for_objects


class NamespaceFactoryFunc(Protocol):
    def __call__(
        self, name: str, extra_metadata: Optional[dict] = None, extra_spec: Optional[dict] = None
    ) -> pykube.Namespace: ...


@pytest.fixture(scope="function")
def namespace_factory_function_scope(kube_cluster: Cluster) -> Iterable[NamespaceFactoryFunc]:
    """Return a new namespace that is deleted once the fixture is disposed. Fixture's scope is 'function'."""
    yield from _namespace_factory_impl(kube_cluster)


@pytest.fixture(scope="module")
def namespace_factory(kube_cluster: Cluster) -> Iterable[NamespaceFactoryFunc]:
    """Return a new namespace that is deleted once the fixture is disposed. Fixture's scope is 'module'."""
    yield from _namespace_factory_impl(kube_cluster)


def _namespace_factory_impl(kube_cluster: Cluster) -> Iterable[NamespaceFactoryFunc]:
    """Return a new namespace that is deleted once the fixture is disposed."""
    created_namespaces: List[pykube.Namespace] = []

    def _namespace_factory(
        name: str,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
    ) -> pykube.Namespace:
        for namespace in created_namespaces:
            if namespace.metadata["name"] == name:
                return namespace

        ns, created = ensure_namespace_exists(kube_cluster.kube_client, name, extra_metadata, extra_spec)
        logger.debug(f"Ensured the namespace '{name}'.")
        if created:
            created_namespaces.append(ns)
        return ns

    yield _namespace_factory

    delete_and_wait_for_objects(kube_cluster.kube_client, pykube.Namespace, created_namespaces)


def _random_ns_name() -> str:
    return f"pytest-{''.join(random.choices(string.ascii_lowercase, k=5))}"  # nosec B311 - non-cryptographic use


@pytest.fixture(scope="function")
def random_namespace_function_scope(namespace_factory: NamespaceFactoryFunc) -> pykube.Namespace:
    """Create and return a random kubernetes namespace that will be deleted at the end of test run.
    Fixture's scope is 'module'."""
    return namespace_factory(_random_ns_name())


@pytest.fixture(scope="module")
def random_namespace(namespace_factory: NamespaceFactoryFunc) -> pykube.Namespace:
    """Create and return a random kubernetes namespace that will be deleted at the end of test run.
    Fixture's scope is 'module'."""
    return namespace_factory(_random_ns_name())



"""Different utilities required over the whole testing lib."""

import logging
import time
from typing import Dict, Any, List, TypeVar, Callable, Type, Optional, Iterable

import pykube.exceptions
from pykube import HTTPClient

from pytest_helm_charts.clusters import Cluster
from pytest_helm_charts.errors import WaitTimeoutError, ObjectStatusError

DEFAULT_DELETE_TIMEOUT_SEC = 120

YamlDict = Dict[str, Any]

logger = logging.getLogger(__name__)

TNS = TypeVar("TNS", bound=pykube.objects.NamespacedAPIObject)
T = TypeVar("T", bound=pykube.objects.APIObject)
FactoryFunc = Callable[..., T]
MetaFactoryFunc = Callable[[pykube.HTTPClient, List[T]], FactoryFunc]


def wait_for_objects_condition(  # noqa: C901
    kube_client: HTTPClient,
    obj_type: Type[T],
    obj_names: List[str],
    objs_namespace: Optional[str],
    obj_condition_func: Callable[[T], bool],
    timeout_sec: int,
    missing_ok: bool,
    failure_condition_func: Optional[Callable[[T], bool]] = None,
) -> List[T]:
    """
    Block until all the kubernetes objects of type `obj_type` pass `obj_condition_fun` or timeout is reached.
        If `objs_namespace` is None, objects are treated as cluster-scope, otherwise as namespace-scope. If
        optional `failure_condition_func` is passed, it is executed when objects are refreshed and if it evaluates to
        `True`, it throws `ObjectStatusError` exception.

    Args:
        kube_client: client to use to connect to the k8s cluster
        obj_type: type of the objects to check; they most be derived from
            [APIObject](pykube.objects.APIObject)
        obj_names: a list of object resource names to check; all the objects must pass `obj_condition_fun`
            for this function to end with success
        objs_namespace: namespace where all the resources should be present (single namespace for all resources).
            If 'None', then it is assumed objects passed are cluster-scope
        timeout_sec: timeout for the call
        obj_condition_func: a function that gets one instance of the resource object of type `obj_type`
            and returns boolean showing whether the object meets the condition or not. The call succeeds
            only if this `obj_condition_fun` returns `True` for every object passed in `obj_names`.
        missing_ok: when `True`, the function ignores that some objects listed in the `obj_names`
            don't exist in k8s API and waits for them to show up; when `False`, an
            [ObjectNotExist](pykube.exceptions.ObjectDoesNotExist) exception is raised.
        failure_condition_func: if not None, then each monitored object is passed to this function. If it
            returns `True`, the `ObjectStatusError` is raised.

    Returns:
        The list of object resources with all the objects listed in `obj_names` included in the list.

    Raises:
        TimeoutError: when timeout is reached.
        pykube.exceptions.ObjectDoesNotExist: when `missing_ok == False` and one of the objects
            listed in `obj_names` can't be found in k8s API
        ObjectStatusError: when `failure_condition_func` is not None and any of the objects in `obj_names`
            returned `True` from this function.

    """
    if len(obj_names) == 0:
        raise ValueError("'obj_names' list can't be empty.")

    retries = 0
    all_ready = False
    matching_objs: List[T] = []
    while retries < timeout_sec:
        response = obj_type.objects(kube_client)
        if objs_namespace:
            response = response.filter(namespace=objs_namespace)
        retries += 1
        matching_objs = []
        for name in obj_names:
            try:
                obj = response.get_by_name(name)
                failed = failure_condition_func and failure_condition_func(obj)
                if failed:
                    raise ObjectStatusError(
                        f"Object's '{obj.namespace}/{obj.name}' status shows failure when waiting "
                        f"for the object's condition to pass."
                    )
                matching_objs.append(obj)
            except pykube.exceptions.ObjectDoesNotExist:
                if missing_ok:
                    pass
                else:
                    raise

        all_ready = len(matching_objs) == len(obj_names) and all(obj_condition_func(obj) for obj in matching_objs)
        if all_ready:
            break
        time.sleep(1)

    if not all_ready:
        raise TimeoutError(f"Error waiting for object of type {obj_type} to match the condition.")

    return matching_objs


def inject_extra(
    cr_dict: YamlDict,
    extra_metadata: Optional[YamlDict] = None,
    extra_spec: Optional[YamlDict] = None,
) -> YamlDict:
    if extra_metadata:
        cr_dict["metadata"].update(extra_metadata)
    if extra_spec:
        cr_dict["spec"].update(extra_spec)
    return cr_dict


def delete_and_wait_for_objects(
    kube_client: HTTPClient,
    obj_type: Type[T],
    objects_to_del: Iterable[T],
    timeout_sec: int = DEFAULT_DELETE_TIMEOUT_SEC,
) -> None:
    """
    For each object in `objects_to_delete`, make an API call to delete it, then wait until the object is gone
    from k8s API server.

    Args:
        kube_client: client to use to connect to the k8s cluster
        obj_type: type of the objects to check; they should be derived from
            [APIObject](pykube.objects.APIObject)
        objects_to_del: iterable of [APIObject](pykube.objects.APIObject) to delete. All objects must be of the same
            specific type.
        timeout_sec: timeout for all the objects in the list to be gone from API server.

    Returns: None

    """
    for kube_object in objects_to_del:
        kube_object.delete()
        obj_name = f"{kube_object.namespace}/{kube_object.name}" if kube_object.namespace else kube_object.name
        logger.debug(f"Deleted object of kind '{obj_type}' named '{obj_name}'.")

    any_exists = True
    times = 0
    while any_exists:
        if times >= timeout_sec:
            raise WaitTimeoutError(f"timeout of {timeout_sec} s exceeded while waiting for objects to be deleted")
        any_exists = False
        for o in objects_to_del:
            objects_method = getattr(obj_type, "objects")
            objects_res = (
                objects_method(kube_client, namespace=o.namespace)
                if "namespace" in o.metadata
                else objects_method(kube_client)
            )
            if objects_res.get_or_none(name=o.name):
                any_exists = True
                time.sleep(1)
                times += 1
                break


def object_factory_helper(
    kube_cluster: Cluster,
    meta_func: MetaFactoryFunc,
    obj_type: Type[T],
    timeout_sec: int = DEFAULT_DELETE_TIMEOUT_SEC,
) -> Iterable[FactoryFunc]:
    created_objects: List[T] = []

    yield meta_func(kube_cluster.kube_client, created_objects)

    delete_and_wait_for_objects(kube_cluster.kube_client, obj_type, created_objects, timeout_sec)



from typing import List, Optional

import pykube
from pykube import Job, HTTPClient

from pytest_helm_charts.utils import wait_for_objects_condition, inject_extra


def _job_complete(job: Job) -> bool:
    complete = (
        "status" in job.obj
        and "conditions" in job.obj["status"]
        and len(job.obj["status"]["conditions"]) > 0
        and job.obj["status"]["conditions"][0]["type"] == "Complete"
    )
    return complete


def wait_for_jobs_to_complete(
    kube_client: HTTPClient, job_names: List[str], jobs_namespace: str, timeout_sec: int, missing_ok: bool = True
) -> List[Job]:
    """
    Block until all the Jobs are complete or timeout is reached.

    Args:
        kube_client: client to use to connect to the k8s cluster
        job_names: a list of Job names to check
        jobs_namespace: namespace where all the Jobs are created (single namespace for all resources)
        timeout_sec: timeout for the call
        missing_ok: when `True`, the function ignores that some of the objects listed in the `job_names`
            don't exist in k8s API and waits for them to show up; when `False`, an
            [ObjectNotFound](pykube.exceptions.ObjectDoesNotExist) exception is raised.

    Returns:
        The list of Job resources with all the objects listed in `job_names` included.

    Raises:
        TimeoutError: when timeout is reached.
        pykube.exceptions.ObjectDoesNotExist: when `missing_ok == False` and one of the objects
            listed in `job_names` can't be found in k8s API

    """
    result = wait_for_objects_condition(
        kube_client, Job, job_names, jobs_namespace, _job_complete, timeout_sec, missing_ok
    )
    return result


def make_job_object(
    kube_client: pykube.HTTPClient,
    name_prefix: str,
    namespace: str,
    command: List[str],
    image: str = "quay.io/giantswarm/busybox:1.32.0",
    restart_policy: str = "OnFailure",
    backoff_limit: int = 6,  # 6 is the default from k8s
    extra_metadata: Optional[dict] = None,
    extra_spec: Optional[dict] = None,
) -> pykube.Job:
    """
    Creates a Job object according to the parameters.

    Args:
        kube_client: client to use to connect to the k8s cluster
        name_prefix: `metadata.generateName` Job's object property that is used to name Pods
        namespace: Namespace to create the Job in
        command: command to run inside the Job pod
        image: container image to use
        restart_policy: Job's restart policy (as in k8s API)
        backoff_limit: Job's backoff limit (as in k8s API)
        extra_metadata: optional dict that will be merged with the 'metadata:' section of the object
        extra_spec: optional dict that will be merged with the 'spec:' section of the object

    Returns:
        Job object. The Job is not sent for creation to API server.
    """
    obj = inject_extra(
        {
            "apiVersion": "batch/v1",
            "kind": "Job",
            "metadata": {"generateName": name_prefix, "namespace": namespace},
            "spec": {
                "backoffLimit": backoff_limit,
                "template": {
                    "spec": {
                        "containers": [
                            {
                                "name": f"{name_prefix}job",
                                "image": image,
                                "command": command,
                            }
                        ],
                        "restartPolicy": restart_policy,
                    },
                },
            },
        },
        extra_metadata,
        extra_spec,
    )
    return pykube.Job(kube_client, obj)


def create_job_and_run_to_completion(
    kube_client: pykube.HTTPClient,
    namespace: str,
    job: pykube.Job,
    timeout_sec: int = 60,
    missing_ok: bool = False,
) -> pykube.Job:
    """
    Creates Job object in k8s and blocks until it is completed.

    Args:
        kube_client: client to use to connect to the k8s cluster
        namespace: Namespace to create the Job in
        job: Job object to run and check.
        timeout_sec: timeout for the call
        missing_ok: when `True`, the function ignores that the Job doesn't yet exist in k8s API
            and waits for it to show up;
            when `False`, an [ObjectNotFound](pykube.exceptions.ObjectDoesNotExist) exception is raised.

    Returns:
        The Job object with refreshed state.

    Raises:
        TimeoutError: when timeout is reached.
        pykube.exceptions.ObjectDoesNotExist: when `missing_ok == False` and the `job` can't be found in k8s API

    """
    job.create()

    wait_for_jobs_to_complete(
        kube_client,
        [job.name],
        namespace,
        timeout_sec,
        missing_ok=missing_ok,
    )

    return job



import logging
from typing import List, Iterable

import pytest
from deprecated import deprecated
from pykube import ConfigMap

from pytest_helm_charts.k8s.fixtures import NamespaceFactoryFunc
from pytest_helm_charts.clusters import Cluster
from pytest_helm_charts.giantswarm_app_platform.app import (
    AppFactoryFunc,
    app_factory_func,
    ConfiguredApp,
    AppCR,
)
from pytest_helm_charts.giantswarm_app_platform.app_catalog import (
    AppCatalogFactoryFunc,
    AppCatalogCR,
    app_catalog_factory_func,
)
from pytest_helm_charts.giantswarm_app_platform.catalog import (
    CatalogFactoryFunc,
    CatalogCR,
    catalog_factory_func,
)
from pytest_helm_charts.utils import object_factory_helper, delete_and_wait_for_objects

logger = logging.getLogger(__name__)


@deprecated(version="0.5.3", reason="Please use `catalog_factory` fixture instead.")
@pytest.fixture(scope="module")
def app_catalog_factory(kube_cluster: Cluster) -> Iterable[AppCatalogFactoryFunc]:
    """
    [Obsolete]
    Please use
    [_catalog_factory](pytest_helm_charts.giantswarm_app_platform.catalog._catalog_factory) instead.

    Return a factory object, that can be used to configure new AppCatalog CRs
    for the 'app-operator' running in the cluster. Fixture's scope is 'module'.
    """
    yield from object_factory_helper(kube_cluster, app_catalog_factory_func, AppCatalogCR)


@pytest.fixture(scope="function")
def catalog_factory_function_scope(
    kube_cluster: Cluster, namespace_factory: NamespaceFactoryFunc
) -> Iterable[CatalogFactoryFunc]:
    """Return a factory object, that can be used to configure new Catalog CRs
    for the 'app-operator' running in the cluster. Fixture's scope is 'function'."""
    yield from _catalog_factory_impl(kube_cluster, namespace_factory)


@pytest.fixture(scope="module")
def catalog_factory(kube_cluster: Cluster, namespace_factory: NamespaceFactoryFunc) -> Iterable[CatalogFactoryFunc]:
    """Return a factory object, that can be used to configure new Catalog CRs
    for the 'app-operator' running in the cluster. Fixture's scope is 'module'."""
    yield from _catalog_factory_impl(kube_cluster, namespace_factory)


def _catalog_factory_impl(
    kube_cluster: Cluster, namespace_factory: NamespaceFactoryFunc
) -> Iterable[CatalogFactoryFunc]:
    created_objects: List[CatalogCR] = []

    yield catalog_factory_func(kube_cluster.kube_client, created_objects, namespace_factory)

    delete_and_wait_for_objects(kube_cluster.kube_client, CatalogCR, created_objects)


@pytest.fixture(scope="module")
def app_factory(
    kube_cluster: Cluster, catalog_factory: CatalogFactoryFunc, namespace_factory: NamespaceFactoryFunc
) -> Iterable[AppFactoryFunc]:
    """Returns a factory function which can be used to install an app using App CR. Fixture's scope is 'module'."""
    yield from _app_factory_impl(kube_cluster, catalog_factory, namespace_factory)


@pytest.fixture(scope="function")
def app_factory_function_scope(
    kube_cluster: Cluster, catalog_factory: CatalogFactoryFunc, namespace_factory: NamespaceFactoryFunc
) -> Iterable[AppFactoryFunc]:
    """Returns a factory function which can be used to install an app using App CR. Fixture's scope is 'module'."""
    yield from _app_factory_impl(kube_cluster, catalog_factory, namespace_factory)


def _app_factory_impl(
    kube_cluster: Cluster, catalog_factory: CatalogFactoryFunc, namespace_factory: NamespaceFactoryFunc
) -> Iterable[AppFactoryFunc]:
    """Returns a factory function which can be used to install an app using App CR."""

    created_apps: List[ConfiguredApp] = []

    yield app_factory_func(kube_cluster.kube_client, catalog_factory, namespace_factory, created_apps)

    apps_to_delete = [a.app for a in created_apps]
    delete_and_wait_for_objects(kube_cluster.kube_client, AppCR, apps_to_delete)
    cms_to_delete = [a.app_cm for a in created_apps if a.app_cm is not None]
    delete_and_wait_for_objects(kube_cluster.kube_client, ConfigMap, cms_to_delete)



import logging
from typing import List, Optional, Protocol, Dict

from pykube import HTTPClient
from pykube.objects import NamespacedAPIObject

from pytest_helm_charts.k8s.fixtures import NamespaceFactoryFunc
from pytest_helm_charts.utils import inject_extra

logger = logging.getLogger(__name__)


class CatalogCR(NamespacedAPIObject):
    version = "application.giantswarm.io/v1alpha1"
    endpoint = "catalogs"
    kind = "Catalog"


class CatalogFactoryFunc(Protocol):
    def __call__(
        self,
        catalog_name: str,
        catalog_namespace: str,
        catalog_url: Optional[str],
        repositories_urls: Optional[List[str]] = None,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
    ) -> CatalogCR: ...


def make_catalog_obj(
    kube_client: HTTPClient,
    catalog_name: str,
    catalog_namespace: str,
    catalog_url: str,
    repositories_urls: Optional[List[str]] = None,
    extra_metadata: Optional[dict] = None,
    extra_spec: Optional[dict] = None,
) -> CatalogCR:
    repositories: List[Dict[str, str]] = []
    if repositories_urls:
        repositories = [{"URL": r, "type": "helm"} for r in repositories_urls]
    else:
        repositories = [{"URL": catalog_url, "type": "helm"}]

    catalog_obj = inject_extra(
        {
            "apiVersion": "application.giantswarm.io/v1alpha1",
            "kind": "Catalog",
            "metadata": {
                "name": catalog_name,
                "namespace": catalog_namespace,
            },
            "spec": {
                "description": "Catalog for testing.",
                "storage": {"URL": catalog_url, "type": "helm"},
                "repositories": repositories,
                "title": catalog_name,
                "logoURL": "https://my-org.github.com/logo.png",
            },
        },
        extra_metadata,
        extra_spec,
    )
    return CatalogCR(kube_client, catalog_obj)


def catalog_factory_func(
    kube_client: HTTPClient, objects: List[CatalogCR], namespace_factory: NamespaceFactoryFunc
) -> CatalogFactoryFunc:
    """Return a factory object, that can be used to configure new Catalog CRs
    for the 'app-operator' running in the cluster"""

    def _catalog_factory(
        catalog_name: str,
        catalog_namespace: str = "default",
        catalog_url: Optional[str] = None,
        repositories_urls: Optional[List[str]] = None,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
    ) -> CatalogCR:
        """A factory function used to create catalogs in the k8s API using Catalog CR.

        Args:
            catalog_name: name of the created Catalog CR. If the Catalog with this name already exists
                in the namespace but the URL is different, it's an error. If the URL and name
                are the same, nothing is done.
            catalog_namespace: namespace to create the Catalog CR in.
            catalog_url: URL of the catalog.
            repositories_urls: A list of additional repositories hosting the same catalog. If left at 'None',
                a single entry including `catalog_url` will be created.
            extra_metadata: any additional fields to be included in the 'metadata' section.
            extra_spec: any additional fields to be included in the 'Spec' section.

        Returns:
            CatalogCR created or found in the k8s API.

        Raises:
            ValueError: if catalog with the same name, but different URL already exists.

        """

        namespace_factory(catalog_namespace)
        if not catalog_url:
            catalog_url = "https://giantswarm.github.io/{}-catalog/".format(catalog_name)
        for c in objects:
            if c.metadata["name"] == catalog_name and c.metadata["namespace"] == catalog_namespace:
                existing_url = c.obj["spec"]["storage"]["URL"]
                if existing_url == catalog_url:
                    return c
                raise ValueError(
                    f"You requested creation of Catalog named {catalog_name} in namespace {catalog_namespace} "
                    f"with URL {catalog_url}, but it was already registered with another URL {existing_url}."
                )

        catalog = make_catalog_obj(
            kube_client, catalog_name, catalog_namespace, catalog_url, repositories_urls, extra_metadata, extra_spec
        )
        objects.append(catalog)
        catalog.create()
        logger.debug(f"Created Catalog '{catalog.namespace}/{catalog.name}'.")
        # TODO: once Catalog CR supports `status` fields, check here that the catalog is present
        return catalog

    return _catalog_factory



import logging
from typing import List, Optional, Protocol

from deprecated import deprecated
from pykube import HTTPClient
from pykube.objects import APIObject

from pytest_helm_charts.utils import inject_extra

logger = logging.getLogger(__name__)


@deprecated(version="0.5.3", reason="Please use `CatalogCR` instead.")
class AppCatalogCR(APIObject):
    version = "application.giantswarm.io/v1alpha1"
    endpoint = "appcatalogs"
    kind = "AppCatalog"


class AppCatalogFactoryFunc(Protocol):
    def __call__(
        self,
        catalog_name: str,
        catalog_url: Optional[str],
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
    ) -> AppCatalogCR: ...


@deprecated(version="0.5.3", reason="Please use `catalog.get_catalog_obj()` instead.")
def make_app_catalog_object(
    kube_client: HTTPClient,
    catalog_name: str,
    catalog_url: str,
    extra_metadata: Optional[dict] = None,
    extra_spec: Optional[dict] = None,
) -> AppCatalogCR:
    app_catalog_cr = inject_extra(
        {
            "apiVersion": "application.giantswarm.io/v1alpha1",
            "kind": "AppCatalog",
            "metadata": {
                "labels": {"app-operator.giantswarm.io/version": "1.0.0", "application.giantswarm.io/catalog-type": ""},
                "name": catalog_name,
            },
            "spec": {
                "description": "Catalog for testing.",
                "storage": {"URL": catalog_url, "type": "helm"},
                "title": catalog_name,
                "logoURL": "https://my-org.github.com/logo.png",
            },
        },
        extra_metadata,
        extra_spec,
    )

    return AppCatalogCR(kube_client, app_catalog_cr)


@deprecated(version="0.5.3", reason="Please use `catalog.catalog_factory_func()` instead.")
def app_catalog_factory_func(
    kube_client: HTTPClient, created_app_catalogs: List[AppCatalogCR]
) -> AppCatalogFactoryFunc:
    """Return a factory object, that can be used to configure new AppCatalog CRs
    for the 'app-operator' running in the cluster"""

    def _app_catalog_factory(
        catalog_name: str,
        catalog_url: Optional[str] = None,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
    ) -> AppCatalogCR:
        """
        [Obsolete] Please use
        [_catalog_factory](pytest_helm_charts.giantswarm_app_platform.catalog._catalog_factory) instead.

        A factory function used to create catalogs in the k8s API using AppCatalog CR.

        Args:
            catalog_name: name of the created AppCatalog CR. If the name already exists and the URL is
                different, it's an error. If the URL and name are the same, nothing is done.
            catalog_url: URL of the catalog.

        Returns:
            AppCatalogCR created or found in the k8s API.

        Raises:
            ValueError: if catalog with the same name, but different URL already exists.

        """
        if not catalog_url:
            catalog_url = "https://giantswarm.github.io/{}-catalog/".format(catalog_name)
        for c in created_app_catalogs:
            if c.metadata["name"] == catalog_name:
                existing_url = c.obj["spec"]["storage"]["URL"]
                if existing_url == catalog_url:
                    return c
                raise ValueError(
                    "You requested creation of AppCatalog named {} with URL {} but it was already registered with URL "
                    "{}".format(catalog_name, catalog_url, existing_url)
                )

        app_catalog = make_app_catalog_object(kube_client, catalog_name, catalog_url, extra_metadata, extra_spec)
        created_app_catalogs.append(app_catalog)
        app_catalog.create()
        logger.debug(f"Created AppCatalog '{app_catalog.name}'.")
        return app_catalog

    return _app_catalog_factory



from typing import Optional, Tuple

import pykube

from pytest_helm_charts.utils import inject_extra


def make_namespace_object(
    kube_client: pykube.HTTPClient,
    namespace_name: str,
    extra_metadata: Optional[dict] = None,
    extra_spec: Optional[dict] = None,
) -> pykube.Namespace:
    """
    Creates a Namespace object in the API server.
    Args:
        kube_client: client to use to connect to the k8s cluster
        namespace_name: a name of the Namespace to ensure
        extra_metadata: optional dict that will be merged with the 'metadata:' section of the object
        extra_spec: optional dict that will be merged with the 'spec:' section of the object

    Returns:
        Namespace resource object.

    """
    obj = inject_extra(
        {
            "apiVersion": "v1",
            "kind": "Namespace",
            "metadata": {
                "name": namespace_name,
            },
        },
        extra_metadata,
        extra_spec,
    )
    return pykube.Namespace(kube_client, obj)


def ensure_namespace_exists(
    kube_client: pykube.HTTPClient,
    namespace_name: str,
    extra_metadata: Optional[dict] = None,
    extra_spec: Optional[dict] = None,
) -> Tuple[pykube.Namespace, bool]:
    """
    Checks if the Namespace exists and creates it if it doesn't
    Args:
        kube_client: client to use to connect to the k8s cluster
        namespace_name: a name of the Namespace to ensure
        extra_metadata: optional dict that will be merged with the 'metadata:' section of the object
        extra_spec: optional dict that will be merged with the 'spec:' section of the object

    Returns:
        Namespace resource object and bool equal True if the namespace was created by this function.

    """
    created = False
    ns = pykube.Namespace.objects(kube_client).get_or_none(name=namespace_name)
    if ns is None:
        ns = make_namespace_object(kube_client, namespace_name, extra_metadata, extra_spec)
        ns.create()
        created = True
    return ns, created



from typing import List

import pykube
from pykube import HTTPClient

from pytest_helm_charts.utils import wait_for_objects_condition


def _daemon_set_ready(ds: pykube.DaemonSet) -> bool:
    complete = (
        "desiredNumberScheduled" in ds.obj["status"]
        and "numberReady" in ds.obj["status"]
        and int(ds.obj["status"]["desiredNumberScheduled"]) == int(ds.obj["status"]["numberReady"])
    )
    return complete


def wait_for_daemon_sets_to_run(
    kube_client: HTTPClient,
    daemon_set_names: List[str],
    daemon_sets_namespace: str,
    timeout_sec: int,
    missing_ok: bool = False,
) -> List[pykube.DaemonSet]:
    """
    Block until all the DaemonSets are running or timeout is reached.

    Args:
        kube_client: client to use to connect to the k8s cluster
        daemon_set_names: a list of DaemonSet names to check
        daemon_sets_namespace: namespace where all the DaemonSets are created (single namespace for all resources)
        timeout_sec: timeout for the call
        missing_ok: when `True`, the function ignores that some of the objects listed in the `daemon_set_names`
            don't exist in k8s API and waits for them to show up; when `False`, an
            [ObjectNotFound](pykube.exceptions.ObjectDoesNotExist) exception is raised.

    Returns:
        The list of DaemonSet resources with all the objects listed in `daemon_set_names` included.

    Raises:
        TimeoutError: when timeout is reached.
        pykube.exceptions.ObjectDoesNotExist: when `missing_ok == False` and one of the objects
            listed in `daemon_set_names` can't be found in k8s API

    """
    result = wait_for_objects_condition(
        kube_client,
        pykube.DaemonSet,
        daemon_set_names,
        daemon_sets_namespace,
        _daemon_set_ready,
        timeout_sec,
        missing_ok=missing_ok,
    )
    return result



import abc
import logging

from pykube.objects import NamespacedAPIObject

logger = logging.getLogger(__name__)

FLUX_CR_READY_TIMEOUT_SEC = 30


class NamespacedFluxCR(NamespacedAPIObject, abc.ABC):
    pass


def flux_cr_ready(flux_obj: NamespacedFluxCR) -> bool:
    has_conditions = "status" in flux_obj.obj and "conditions" in flux_obj.obj["status"]
    if not has_conditions:
        return False
    conditions_count = len(flux_obj.obj["status"]["conditions"])
    if conditions_count == 0:
        return False
    if conditions_count > 1:
        logging.warning(
            f"Found '{conditions_count}' status conditions when expecting just 1. Using only the first one on the list."
        )
    condition = flux_obj.obj["status"]["conditions"][0]
    return condition["type"] == "Ready" and condition["status"] == "True"



import logging
from typing import Protocol, Optional, Any, List, Dict

from pykube import HTTPClient

from pytest_helm_charts.k8s.fixtures import NamespaceFactoryFunc
from pytest_helm_charts.flux.utils import NamespacedFluxCR, flux_cr_ready
from pytest_helm_charts.utils import wait_for_objects_condition, inject_extra


logger = logging.getLogger(__name__)


class KustomizationCR(NamespacedFluxCR):
    version = "kustomize.toolkit.fluxcd.io/v1beta2"
    endpoint = "kustomizations"
    kind = "Kustomization"


class KustomizationFactoryFunc(Protocol):
    def __call__(
        self,
        name: str,
        namespace: str,
        prune: bool,
        interval: str,
        repo_path: str,
        git_repository_name: str,
        timeout: str,
        service_account_name: Optional[str] = None,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
        wait_timeout_sec: int = 30,
    ) -> KustomizationCR: ...


def kustomization_factory_func(
    kube_client: HTTPClient,
    namespace_factory: NamespaceFactoryFunc,
    created_kustomizations: List[KustomizationCR],
) -> KustomizationFactoryFunc:
    """Return a factory object, that can be used to create a new Kustomization CRs"""

    def _kustomization_factory(
        name: str,
        namespace: str,
        prune: bool,
        interval: str,
        repo_path: str,
        git_repository_name: str,
        timeout: str,
        service_account_name: Optional[str] = None,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
        wait_timeout_sec: int = 30,
    ) -> KustomizationCR:
        """A factory function used to create Flux Kustomizations.
        Args:
            name: name of the created Kustomization CR.
            namespace: namespace to create the Kustomization CR in.
            prune: enables garbage collection.
            interval: the interval at which to reconcile the Kustomization.
            repo_path: path to the directory containing the kustomization.yaml file.
            git_repository_name: name of the GitRepository object to get this kustomization from.
            timeout: timeout for apply and health checking operations.
            service_account_name: the name of the Kubernetes service account to impersonate
                when reconciling this Kustomization.
            extra_metadata: a dictionary of any additional attributes to put directly into "metadata"
                part of the object
            extra_spec: a dictionary of any additional attributes to put directly into "spec"
                part of the object
            wait_timeout_sec: How long to wait for the HelmRelease to be ready.
        Returns:
            KustomizationCR created or found in the k8s API.
        Raises:
            ValueError: if object with the same name already exists.
        """
        for k in created_kustomizations:
            if k.metadata["name"] == name and k.metadata["namespace"] == namespace:
                return k

        namespace_factory(namespace)
        kustomization = make_kustomization_obj(
            kube_client,
            name,
            namespace,
            prune,
            interval,
            repo_path,
            git_repository_name,
            timeout,
            service_account_name,
            extra_metadata=extra_metadata,
            extra_spec=extra_spec,
        )
        created_kustomizations.append(kustomization)
        kustomization.create()
        logger.debug(f"Created Flux Kustomization '{kustomization.namespace}/{kustomization.name}'.")
        wait_for_kustomizations_to_be_ready(kube_client, [name], namespace, wait_timeout_sec, missing_ok=True)
        return kustomization

    return _kustomization_factory


def make_kustomization_obj(
    kube_client: HTTPClient,
    name: str,
    namespace: str,
    prune: bool,
    interval: str,
    repo_path: str,
    git_repository_name: str,
    timeout: str,
    service_account_name: Optional[str] = None,
    extra_metadata: Optional[dict] = None,
    extra_spec: Optional[dict] = None,
) -> KustomizationCR:
    cr: Dict[str, Any] = {
        "apiVersion": KustomizationCR.version,
        "kind": KustomizationCR.kind,
        "metadata": {
            "name": name,
            "namespace": namespace,
        },
        "spec": {
            "prune": prune,
            "interval": interval,
            "path": repo_path,
            "sourceRef": {
                "kind": "GitRepository",
                "name": git_repository_name,
            },
            "timeout": timeout,
        },
    }
    if service_account_name:
        cr["spec"]["serviceAccountName"] = service_account_name

    cr = inject_extra(cr, extra_metadata, extra_spec)
    return KustomizationCR(kube_client, cr)


def wait_for_kustomizations_to_be_ready(
    kube_client: HTTPClient,
    kustomization_names: List[str],
    kustomization_namespace: str,
    timeout_sec: int,
    missing_ok: bool = False,
) -> List[KustomizationCR]:
    """Block until all Kustomization objects in `kustomization_names` have status 'Ready'."""
    objects = wait_for_objects_condition(
        kube_client,
        KustomizationCR,
        kustomization_names,
        kustomization_namespace,
        flux_cr_ready,
        timeout_sec,
        missing_ok,
    )
    return objects



from _pytest.config.argparsing import Parser

from pytest_helm_charts.fixtures import (  # noqa: F401
    chart_path,
    chart_version,
    test_extra_info,
    cluster_type,
    kube_cluster,
    kube_config,
    values_file_path,
    get_cmd_line_option_name_from_env_var,
    CMD_VAR_TEST_EXTRA_INFO,
    ENV_VAR_CHART_PATH,
    ENV_VAR_CHART_VERSION,
    ENV_VAR_CLUSTER_TYPE,
    ENV_VAR_CLUSTER_VERSION,
    ENV_VAR_KUBE_CONFIG,
    ENV_VAR_APP_CONFIG_PATH,
)
from pytest_helm_charts.flux.fixtures import (  # noqa: F401
    flux_deployments,
    kustomization_factory,
    kustomization_factory_function_scope,
    git_repository_factory,
    git_repository_factory_function_scope,
    helm_repository_factory,
    helm_repository_factory_function_scope,
    helm_release_factory,
    helm_release_factory_function_scope,
)
from pytest_helm_charts.giantswarm_app_platform.apps.http_testing import (  # noqa: F401
    gatling_app_factory,
    stormforger_load_app_factory,
)
from pytest_helm_charts.giantswarm_app_platform.fixtures import (  # noqa: F401
    app_catalog_factory,
    app_factory,
    app_factory_function_scope,
    catalog_factory,
    catalog_factory_function_scope,
)
from pytest_helm_charts.k8s.fixtures import (  # noqa: F401
    namespace_factory,
    namespace_factory_function_scope,
    random_namespace,
    random_namespace_function_scope,
)


def _get_cmd_line_option_full_name(env_var_name: str) -> str:
    cmd_name = get_cmd_line_option_name_from_env_var(env_var_name)
    cmd_name = cmd_name.replace("_", "-")
    return "--" + cmd_name


def pytest_addoption(parser: Parser) -> None:
    group = parser.getgroup("pytest-helm-charts")
    group.addoption(
        _get_cmd_line_option_full_name(ENV_VAR_CHART_PATH), action="store", help="The path to a helm chart under test."
    )
    group.addoption(
        _get_cmd_line_option_full_name(ENV_VAR_CHART_VERSION),
        action="store",
        help="Override chart version for the chart under test.",
    )
    group.addoption(
        _get_cmd_line_option_full_name(ENV_VAR_CLUSTER_TYPE),
        action="store",
        help="Pass information about cluster type being used for tests.",
    )
    group.addoption(
        _get_cmd_line_option_full_name(ENV_VAR_CLUSTER_VERSION),
        action="store",
        help="Pass information about k8s version being used for tests.",
    )
    group.addoption(
        _get_cmd_line_option_full_name(ENV_VAR_KUBE_CONFIG),
        action="store",
        help="The path to 'kube config' file.",
    )
    group.addoption(
        _get_cmd_line_option_full_name(ENV_VAR_APP_CONFIG_PATH),
        action="store",
        help="Path to the values file used for testing the chart.",
    )
    group.addoption(
        "--" + CMD_VAR_TEST_EXTRA_INFO,
        action="store",
        help="Pass any additional info about the test in the 'key1=val1,key2=val2' format",
    )



"""This module defines fixtures for testing Helm Charts."""

import logging
import os
import sys
from typing import Iterable, Dict, Mapping

import pytest
from _pytest.config import Config

from pytest_helm_charts.clusters import ExistingCluster, Cluster

logger = logging.getLogger(__name__)

ENV_VAR_CHART_PATH = "ATS_CHART_PATH"
ENV_VAR_CHART_VERSION = "ATS_CHART_VERSION"
ENV_VAR_CLUSTER_TYPE = "ATS_CLUSTER_TYPE"
ENV_VAR_CLUSTER_VERSION = "ATS_CLUSTER_VERSION"
ENV_VAR_APP_CONFIG_PATH = "ATS_APP_CONFIG_FILE_PATH"
ENV_VAR_KUBE_CONFIG = "KUBECONFIG"
ENV_VAR_ATS_EXTRA_PREFIX = "ATS_EXTRA_"
CMD_VAR_TEST_EXTRA_INFO = "test_extra_info"


def get_cmd_line_option_name_from_env_var(env_var_name: str) -> str:
    cmd_name = env_var_name.lower()
    if cmd_name.startswith("ats_"):
        cmd_name = cmd_name[4:]
    return cmd_name


def _load_mandatory_config_option(pytestconfig: Config, env_var_name: str) -> str:
    cmd_name = get_cmd_line_option_name_from_env_var(env_var_name)
    if pytestconfig.getoption(cmd_name):
        return pytestconfig.getoption(cmd_name)
    if env_var_name in os.environ and len(os.environ[env_var_name]) > 0:
        return os.environ[env_var_name]
    raise Exception(
        f"Environment variable '{env_var_name}' (or relevant cmd line option, see help '-h') needed "
        "by a fixture, but not set."
    )


def _load_optional_config_option(pytestconfig: Config, env_var_name: str) -> str:
    cmd_name = get_cmd_line_option_name_from_env_var(env_var_name)
    if pytestconfig.getoption(cmd_name):
        return pytestconfig.getoption(cmd_name)
    return os.environ[env_var_name] if env_var_name in os.environ else ""


def _parse_cmd_opt_extra_info(info: str) -> Dict[str, str]:
    pairs = list(filter(None, info.split(",")))
    res_dict: Dict[str, str] = {}
    for pair in pairs:
        k, v = list(filter(None, pair.split("=")))
        res_dict[k.lower()] = v
    return res_dict


def _filter_extra_info_from_mapping(extra: Mapping[str, str]) -> Dict[str, str]:
    return {
        k[len(ENV_VAR_ATS_EXTRA_PREFIX) :].lower(): v
        for k, v in extra.items()
        if k.startswith(ENV_VAR_ATS_EXTRA_PREFIX)
    }


@pytest.fixture(scope="module")
def chart_path(pytestconfig: Config) -> str:
    """Return a path to the chart under test (from command line argument)."""
    return _load_optional_config_option(pytestconfig, ENV_VAR_CHART_PATH)


@pytest.fixture(scope="module")
def chart_version(pytestconfig: Config) -> str:
    """Return a value that needs to be used as chart version override (from command line argument)."""
    return _load_optional_config_option(pytestconfig, ENV_VAR_CHART_VERSION)


@pytest.fixture(scope="module")
def cluster_type(pytestconfig: Config) -> str:
    """Return a type of cluster used for testing (from command line argument)."""
    return _load_optional_config_option(pytestconfig, ENV_VAR_CLUSTER_TYPE)


@pytest.fixture(scope="module")
def cluster_version(pytestconfig: Config) -> str:
    """Return a type of cluster used for testing (from command line argument)."""
    return _load_optional_config_option(pytestconfig, ENV_VAR_CLUSTER_VERSION)


@pytest.fixture(scope="module")
def values_file_path(pytestconfig: Config) -> str:
    """Return a path to the yaml file that needs to be used to configure chart under test
    (from command line argument).
    """
    return _load_optional_config_option(pytestconfig, ENV_VAR_APP_CONFIG_PATH)


@pytest.fixture(scope="module")
def kube_config(pytestconfig: Config) -> str:
    """Return a path to the kube.config file that points to a running cluster with app
    catalog platform tools already installed."""
    return _load_mandatory_config_option(pytestconfig, ENV_VAR_KUBE_CONFIG)


@pytest.fixture(scope="module")
def test_extra_info(pytestconfig: Config) -> Dict[str, str]:
    """Return an optional dict of variable names and values passed to the test using either
    the `--extra-test-info` cmd line option or env vars prefixed with 'ATS_EXTRA_'."""
    from_env = _filter_extra_info_from_mapping(os.environ)
    if pytestconfig.getoption(CMD_VAR_TEST_EXTRA_INFO):
        from_cmd = _parse_cmd_opt_extra_info(pytestconfig.getoption(CMD_VAR_TEST_EXTRA_INFO))
        from_env.update(from_cmd)
    return from_env


@pytest.fixture(scope="module")
def kube_cluster(
    kube_config: str,
) -> Iterable[Cluster]:
    """Return a ready Cluster object, which can already be used in test to connect
    to the cluster. Specific implementation used to provide the cluster depends
    on the '--cluster-type' command line option."""
    cluster = ExistingCluster(kube_config)

    cluster.create()
    logger.debug("Cluster connection configured")
    yield cluster

    # noinspection PyBroadException
    try:
        cluster.destroy()
        logger.debug("Cluster connection released")
    except Exception:
        exc = sys.exc_info()
        logger.error(f"Error of type {exc[0]} when releasing cluster. Value: {exc[1]}\nStacktrace:\n{exc[2]}")



import logging
from typing import Protocol, Optional, Any, List, Dict

from pykube import HTTPClient

from pytest_helm_charts.k8s.fixtures import NamespaceFactoryFunc
from pytest_helm_charts.flux.utils import NamespacedFluxCR, flux_cr_ready
from pytest_helm_charts.utils import wait_for_objects_condition, inject_extra


logger = logging.getLogger(__name__)


class HelmRepositoryCR(NamespacedFluxCR):
    version = "source.toolkit.fluxcd.io/v1beta1"
    endpoint = "helmrepositories"
    kind = "HelmRepository"


class HelmRepositoryFactoryFunc(Protocol):
    def __call__(
        self,
        name: str,
        namespace: str,
        interval: str,
        repo_url: str,
        secret_ref_name: Optional[str] = None,
        timeout: Optional[str] = None,
        suspend: bool = False,
        pass_credentials: bool = False,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
        wait_timeout_sec: int = 30,
    ) -> HelmRepositoryCR: ...


def helm_repository_factory_func(
    kube_client: HTTPClient,
    namespace_factory: NamespaceFactoryFunc,
    created_helm_repositories: List[HelmRepositoryCR],
) -> HelmRepositoryFactoryFunc:
    """Return a factory object, that can be used to create a new HelmRepository CRs"""

    def _helm_repository_factory(
        name: str,
        namespace: str,
        interval: str,
        repo_url: str,
        secret_ref_name: Optional[str] = None,
        timeout: Optional[str] = None,
        suspend: bool = False,
        pass_credentials: bool = False,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
        wait_timeout_sec: int = 30,
    ) -> HelmRepositoryCR:
        """A factory function used to create Flux HelmRepository.
        Args:
            name: name of the created Kustomization CR.
            namespace: namespace to create the Kustomization CR in.
            interval: the interval at which to check for repository updates.
            repo_url: the repository URL, can be a HTTP/S or SSH address.
            secret_ref_name: the secret name containing the Git credentials.
                For HTTPS repositories the secret must contain username and password fields.
                For SSH repositories the secret must contain identity, identity.pub and known_hosts fields.
            timeout: The timeout of index downloading, defaults to 60s.
            suspend: This flag tells the controller to suspend the reconciliation of this source.
            pass_credentials: PassCredentials allows the credentials from the SecretRef to be passed on to
                a host that does not match the host as defined in URL.
            extra_metadata: a dictionary of any additional attributes to put directly into "metadata"
                part of the object
            extra_spec: a dictionary of any additional attributes to put directly into "spec"
                part of the object
            wait_timeout_sec: How long to wait for the HelmRelease to be ready.
        Returns:
            HelmRepository created or found in the k8s API.
        Raises:
            ValueError: if object with the same name already exists.
        """
        for hr in created_helm_repositories:
            if hr.metadata["name"] == name and hr.metadata["namespace"] == namespace:
                return hr

        namespace_factory(namespace)
        helm_repository = make_helm_repository_obj(
            kube_client,
            name,
            namespace,
            interval,
            repo_url,
            secret_ref_name,
            timeout,
            suspend,
            pass_credentials,
            extra_metadata=extra_metadata,
            extra_spec=extra_spec,
        )
        created_helm_repositories.append(helm_repository)
        helm_repository.create()
        logger.debug(f"Created Flux HelmRepository '{helm_repository.namespace}/{helm_repository.name}'.")
        wait_for_helm_repositories_to_be_ready(kube_client, [name], namespace, wait_timeout_sec, missing_ok=True)
        return helm_repository

    return _helm_repository_factory


def make_helm_repository_obj(
    kube_client: HTTPClient,
    name: str,
    namespace: str,
    interval: str,
    repo_url: str,
    secret_ref_name: Optional[str] = None,
    timeout: Optional[str] = None,
    suspend: bool = False,
    pass_credentials: bool = False,
    extra_metadata: Optional[dict] = None,
    extra_spec: Optional[dict] = None,
) -> HelmRepositoryCR:
    cr: Dict[str, Any] = {
        "apiVersion": HelmRepositoryCR.version,
        "kind": HelmRepositoryCR.kind,
        "metadata": {
            "name": name,
            "namespace": namespace,
        },
        "spec": {
            "url": repo_url,
            "passCredentials": pass_credentials,
            "interval": interval,
            "suspend": suspend,
        },
    }
    if secret_ref_name:
        cr["spec"]["secretRef"] = {
            "name": secret_ref_name,
        }
    if timeout:
        cr["spec"]["timeout"] = timeout

    cr = inject_extra(cr, extra_metadata, extra_spec)
    return HelmRepositoryCR(kube_client, cr)


def wait_for_helm_repositories_to_be_ready(
    kube_client: HTTPClient,
    helm_repo_names: List[str],
    helm_repo_namespace: str,
    timeout_sec: int,
    missing_ok: bool = False,
) -> List[HelmRepositoryCR]:
    """Block until all Helm Repository objects in `helm_repo_names` have status 'Ready'."""
    objects = wait_for_objects_condition(
        kube_client,
        HelmRepositoryCR,
        helm_repo_names,
        helm_repo_namespace,
        flux_cr_ready,
        timeout_sec,
        missing_ok,
    )
    return objects



import logging
from typing import Protocol, Optional, Any, List, Dict

from pykube import HTTPClient

from pytest_helm_charts.k8s.fixtures import NamespaceFactoryFunc
from pytest_helm_charts.flux.utils import NamespacedFluxCR, flux_cr_ready
from pytest_helm_charts.utils import wait_for_objects_condition, inject_extra


logger = logging.getLogger(__name__)


class GitRepositoryCR(NamespacedFluxCR):
    version = "source.toolkit.fluxcd.io/v1beta1"
    endpoint = "gitrepositories"
    kind = "GitRepository"


class GitRepositoryFactoryFunc(Protocol):
    def __call__(
        self,
        name: str,
        namespace: str,
        interval: str,
        repo_url: str,
        repo_branch: str = "master",
        secret_ref_name: Optional[str] = None,
        ignore_pattern: Optional[str] = None,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
        wait_timeout_sec: int = 30,
    ) -> GitRepositoryCR: ...


def git_repository_factory_func(
    kube_client: HTTPClient,
    namespace_factory: NamespaceFactoryFunc,
    created_git_repositories: List[GitRepositoryCR],
) -> GitRepositoryFactoryFunc:
    """Return a factory object, that can be used to create a new GitRepository CRs"""

    def _git_repository_factory(
        name: str,
        namespace: str,
        interval: str,
        repo_url: str,
        repo_branch: str = "master",
        secret_ref_name: Optional[str] = None,
        ignore_pattern: Optional[str] = None,
        extra_metadata: Optional[dict] = None,
        extra_spec: Optional[dict] = None,
        wait_timeout_sec: int = 30,
    ) -> GitRepositoryCR:
        """A factory function used to create Flux GitRepository.
        Args:
            name: name of the created GitRepository CR.
            namespace: namespace to create the GitRepository CR in.
            interval: the interval at which to check for repository updates.
            repo_url: the repository URL, can be a HTTP/S or SSH address.
            repo_branch: branch of the repo to use.
            secret_ref_name: the secret name containing the Git credentials.
                For HTTPS repositories the secret must contain username and password fields.
                For SSH repositories the secret must contain identity, identity.pub and known_hosts fields.
            ignore_pattern: Ignore overrides the set of excluded patterns in the .sourceignore format
                (which is the same as .gitignore). If not provided, a default will be used,
                consult the documentation for your version to find out what those are.
            extra_metadata: a dictionary of any additional attributes to put directly into "metadata"
                part of the object
            extra_spec: a dictionary of any additional attributes to put directly into "spec"
                part of the object
            wait_timeout_sec: How long to wait for the HelmRelease to be ready.
        Returns:
            GitRepositoryCR created or found in the k8s API.
        Raises:
            ValueError: if object with the same name already exists.
        """
        for gr in created_git_repositories:
            if gr.metadata["name"] == name and gr.metadata["namespace"] == namespace:
                return gr

        namespace_factory(namespace)
        git_repository = make_git_repository_obj(
            kube_client,
            name,
            namespace,
            interval,
            repo_url,
            repo_branch,
            secret_ref_name,
            ignore_pattern,
            extra_metadata=extra_metadata,
            extra_spec=extra_spec,
        )
        created_git_repositories.append(git_repository)
        git_repository.create()
        logger.debug(f"Created Flux GitRepository '{git_repository.namespace}/{git_repository.name}'.")
        wait_for_git_repositories_to_be_ready(kube_client, [name], namespace, wait_timeout_sec, missing_ok=True)
        return git_repository

    return _git_repository_factory


def make_git_repository_obj(
    kube_client: HTTPClient,
    name: str,
    namespace: str,
    interval: str,
    repo_url: str,
    repo_branch: str = "master",
    secret_ref_name: Optional[str] = None,
    ignore_pattern: Optional[str] = None,
    extra_metadata: Optional[dict] = None,
    extra_spec: Optional[dict] = None,
) -> GitRepositoryCR:
    cr: Dict[str, Any] = {
        "apiVersion": GitRepositoryCR.version,
        "kind": GitRepositoryCR.kind,
        "metadata": {
            "name": name,
            "namespace": namespace,
        },
        "spec": {
            "interval": interval,
            "url": repo_url,
            "ref": {
                "branch": repo_branch,
            },
        },
    }
    if secret_ref_name:
        cr["spec"]["secretRef"] = {
            "name": secret_ref_name,
        }
    if ignore_pattern:
        cr["spec"]["ignore"] = ignore_pattern

    cr = inject_extra(cr, extra_metadata, extra_spec)
    return GitRepositoryCR(kube_client, cr)


def wait_for_git_repositories_to_be_ready(
    kube_client: HTTPClient,
    git_repo_names: List[str],
    git_repo_namespace: str,
    timeout_sec: int,
    missing_ok: bool = False,
) -> List[GitRepositoryCR]:
    """Block until all Git Repository objects in `git_repo_names` have status 'Ready'."""
    objects = wait_for_objects_condition(
        kube_client,
        GitRepositoryCR,
        git_repo_names,
        git_repo_namespace,
        flux_cr_ready,
        timeout_sec,
        missing_ok,
    )
    return objects
